qapply <- function(X, FUN, fargs=NULL, commonData=NULL, workDir=getwd(),
                   nCores=as.numeric(.getResources()), global=FALSE,
                   all.names=FALSE, nPerPacket=NULL,
                   tag=deparse(substitute(X)), internalize=TRUE,
                   clearWd=TRUE, parSeed=NULL, lib=NULL)
{
  FUN <- match.fun(FUN)
  results <- NULL
  if (!is.vector(X) || is.object(X))
    X <- as.list(X)
  if (is.null(nPerPacket)) nPerPacket=.nPerPacket(nr=length(X), nc=nCores)
  nPerPacket <- nPerPacket[ !is.na(nPerPacket) & !is.null(nPerPacket) & (nPerPacket!=0) ]
  if (sum(nPerPacket) != length(X)){
    print("Specify nPerPacket as an array whose sum matches nrow(X)")
  }else{
    ## Prevent errors from sloppy tag usage:
    tag <- gsub("[/:*?\"'<>|]", "_", tag)
    if(length(list.dirs(file.path(workDir, "out", tag))) > 0) qclearWd(workDir, tag)
    oldDir <- getwd()
    setwd(workDir)
    if(is.null(commonData) & global){ commonData=as.list(.GlobalEnv, all.names) }
    commonEnv <- list( commonData=commonData, X=X, FUN=FUN, fargs=eval(fargs), packages=(.packages()) )
    if(is.null(lib)){ libPaths = .libPaths() }else{ libPaths=lib }
    save( libPaths, file=paste(tag, "libs.Rdata", sep="-"))
    save( commonEnv, file=paste(tag, "common.Rdata", sep="-"))
    ## On NFS shares, sometimes these take a while to propogate...  Pass the md5 sum to qsub
    library(tools)
    md5lib <- md5sum(paste(tag,"libs.Rdata",sep="-"))
    md5common <- md5sum(paste(tag,"common.Rdata",sep="-"))
    
    unlink("qcontrol.R")  
    funScript <- system.file("qcontrol.R", package="qapply", mustWork=TRUE)
    file.copy(funScript, "qcontrol.R")
    if( is.null(parSeed) ){ parSeed <- rep(1111, 6) }
    if( length(parSeed)!=6 ){ stop("parSeed must be six integers")  }
    for(i in 1:length(nPerPacket)){
      rowBegin <- ifelse( i==1, 1, sum(nPerPacket[1:(i-1)])+1 )
      rowEnd <- sum(nPerPacket[1:i])
      qsub("R --no-save --vanilla", tag=tag, rowBegin=rowBegin, rowEnd=rowEnd,
           cmdArgs= c("--args", tag, rowBegin, rowEnd, i, length(nPerPacket), parSeed, md5lib, md5common, "< qcontrol.R"))
    }
    if(internalize){
      qwait(tag)
      results <- qinternalize(file.path("out", tag))
      if(length(results) != length(X)){warning("Some jobs did not complete, length(x)!=length(results).  Check error logs with clearWd=F for more information") }      
      if(clearWd){
        qclearWd(workDir,tag)
      }
    }
  }
  setwd(oldDir)
  results
}
